@@ -5,30 +5,25 @@ import uuid
 import os
 import json
 
-from typing import Any, Callable, Dict, List, Optional, Tuple
+from typing import Any, Dict, List, Optional, Tuple
 from contextlib import contextmanager
 
 from inowfaasutils.comm.pubsub import PubSubClient
 
 from ..misc.dataclass_helper import asdict
-from ..callback.base import BaseCallback
-from ..callback import get_callback, get_err_callback
 from ..misc.enum import FaasOpState
 from ..misc.model import Request
-from ..misc.singleton import Singleton
 from .model import FaasError, FaasJob, FaasJobTrigger
 from ..storage.firestore import FireStoreClient
 
 from google.cloud.firestore import (
-    DocumentSnapshot,
     DocumentReference,
-    CollectionReference,
 )
 
 _INIT_JOBS_TOTAL = 1
 
 
-class FaasJobManager(metaclass=Singleton):
+class FaasJobManager:
     """Job completion metadata manager. It is a singleton class used for storing information of nested
     executed FaaS operations, and help to implement architectures like Fan-in and Fan-out.
 
@@ -46,30 +41,25 @@ class FaasJobManager(metaclass=Singleton):
     Once all related jobs are finished, end_date is updated with epoch representation (in seconds)
     """
 
-    firestore: FireStoreClient
-    job_collection_name: str
-    """job collection name in firestore"""
-    remaining_job_collection_name: str
-    """remaining job metadata collection name in firestore"""
-    failed_job_collection_name: str
-    """failed job metadata collection name in firestore"""
-    job_id: str
-    op_id: str
-    has_parent: bool
-    faas_trigger_queue: List[FaasJobTrigger]
-    job_parent_idx_list: Optional[List[int]]
-
     def __init__(self):
+        self.job_collection_name: str = ""
+        """job collection name in firestore"""
+        self.remaining_job_collection_name: str = ""
+        """remaining job metadata collection name in firestore"""
+        self.failed_job_collection_name: str = ""
+        """failed job metadata collection name in firestore"""
+
         self._init_envs()
-        self.firestore = FireStoreClient(self.GC_PROJECT_ID)
+        self.firestore: FireStoreClient = FireStoreClient(self.GC_PROJECT_ID)
         self.new_jobs_cnt = 0
         self.diff_increment = 0
-        self.job_id = None
-        self.op_id = None
-        self.process_id = None
-        self.has_parent = False
-        self.faas_trigger_queue = []
-        self.job_parent_idx_list = None
+        self.job_id: Optional[str] = None
+        self.op_id: Optional[str] = None
+        self.faas_trigger_queue: List[FaasJobTrigger] = []
+        self.job_parent_idx_list: List[int] = []
+        self.job_done_collection: Optional[str] = None
+        self.user_id = None
+        self.username = None
 
     def _init_envs(self):
         """Init all environment variables needed to initialize this class
@@ -105,32 +95,6 @@ class FaasJobManager(metaclass=Singleton):
         else:
             raise KeyError("GC_PROJECT_ID not found in environment variables")
 
-    def _end_job(self, req: Request, job_data: FaasJob, last_job: bool = False):
-        """Ends job by adding end_date, and runs callback if any is defined
-
-        Args:
-            req (Request): base Faas Job request
-            job_data (FaasJob): Job metadata
-            last_job (bool, optional): True if is the last job executed in
-            the execution chain. Defaults to False.
-        """
-
-        __end_job: Callable[[DocumentSnapshot], None]
-        if req.callback_type and last_job:
-            callback = get_callback(req)
-
-            @callback.add_callback
-            def __end_job(job_data: FaasJob):
-                job_data.end_date = self._epoch_now()
-
-        else:
-
-            def __end_job(job_data: FaasJob):
-                job_data.end_date = self._epoch_now()
-                return job_data
-
-        return __end_job(job_data)
-
     def _generate_faas_error(self, job_name: str, err: Exception) -> FaasError:
         """Generate FaaS error metadata representation
 
@@ -152,19 +116,6 @@ class FaasJobManager(metaclass=Singleton):
             exception_line=err.__traceback__.tb_lineno,
         )
 
-    def _send_err_callback(
-        self, err_data: FaasError, err_callback: Optional[BaseCallback]
-    ):
-        """Send error to a callback function
-
-        Args:
-            err_data (FaasError): FaaS error metadata representation
-            err_callback (Optional[BaseCallback]): Callback function to be run if a
-            job fails
-        """
-
-        err_callback.callback_function(FaasError.Schema().dump(err_data))
-
     def _trigger_faas_queue(self):
         """Enqueue all FaaS calls in the corresponding topics"""
 
@@ -187,7 +138,6 @@ class FaasJobManager(metaclass=Singleton):
         req: Request,
         curr_state: FaasOpState,
         err: Exception = None,
-        err_callback: Optional[BaseCallback] = None,
     ):
         """Action after job has been executed
 
@@ -196,8 +146,6 @@ class FaasJobManager(metaclass=Singleton):
             req (Request): base Faas Job request
             curr_state (FaasOpState): current state of job execution
             err (Exception, optional): Exception raised on execution. Defaults to None.
-            err_callback (Optional[BaseCallback], optional): Callback function to be run if a
-            job fails. Defaults to None.
         """
 
         state = FaasOpState.SCCS if curr_state != FaasOpState.ERR else FaasOpState.ERR
@@ -213,11 +161,12 @@ class FaasJobManager(metaclass=Singleton):
         )
 
         job_data: FaasJob
-        if self.has_parent:
+        upsert_job_id: str
+        if len(self.job_parent_idx_list) > 0:
             parent_doc_ref = self.firestore.get_document_ref(
                 self.job_collection_name,
                 self.job_id,
-                collection_name=self.job_collection_name,
+                shard_collection_name=self.job_collection_name,
                 shard_idx_list=self.job_parent_idx_list[:-1],
             )
             job_data = FaasJob.Schema().load(
@@ -226,31 +175,28 @@ class FaasJobManager(metaclass=Singleton):
                     self.job_id,
                     "total_jobs",
                     self.new_jobs_cnt,
-                    collection_name=self.job_collection_name,
+                    shard_collection_name=self.job_collection_name,
                     shard_idx_list=self.job_parent_idx_list,
                 ).to_dict()
             )
+            upsert_job_id = str(self.job_parent_idx_list[-1]) 
         else:
             parent_doc_ref = None
             root_doc.total_jobs += self.new_jobs_cnt
             job_data = root_doc
+            upsert_job_id = self.job_id
 
         job_data.state = state
 
         if state == FaasOpState.ERR:
             err_data = self._generate_faas_error(job_name, err)
             self._insert_failed(self.job_id, self.op_id, err_data)
-            if root_doc.ended is None:
-                root_doc.ended = True
-                self._upsert_job(data=root_doc)
-                if err_callback is not None:
-                    self._send_err_callback(err_data, err_callback)
+            self.finish_job(req, root_doc, FaasOpState.ERR)
         elif len(self.faas_trigger_queue) > 0:
             self._trigger_faas_queue()
 
         self._delete_remaining(self.job_id, self.op_id)
 
-        last_job = False
         if (
             not root_doc.ended
             and not self.firestore.has_documents(
@@ -260,14 +206,10 @@ class FaasJobManager(metaclass=Singleton):
                 self.failed_job_collection_name, root_doc_ref
             )
         ):
-            root_doc.ended = True
-            last_job = True
-            if parent_doc_ref is not None:
-                self._upsert_job(data=root_doc)
+            self.finish_job(req, root_doc, FaasOpState.SCCS)
 
-        self._end_job(req, job_data, last_job)
-
-        self._upsert_job(data=job_data, parent_document=parent_doc_ref)
+        job_data.end_date = self._epoch_now()
+        self._upsert_job(data=job_data, job_id=upsert_job_id, parent_document=parent_doc_ref)
 
     def _get_or_default_job(
         self,
@@ -311,6 +253,7 @@ class FaasJobManager(metaclass=Singleton):
         data: FaasJob = None,
         name: str = None,
         args: Any = None,
+        job_id: str = None,
         parent_document: Optional[DocumentReference] = None,
     ) -> Tuple[str, FaasJob]:
         """Upsert faas job execution metadata into FireStore `job` collection
@@ -330,13 +273,13 @@ class FaasJobManager(metaclass=Singleton):
         job_dict = asdict(job)
         self.firestore.upsert(
             self.job_collection_name,
-            self.job_id,
+            job_id,
             job_dict,
             parent_document=parent_document,
         )
-        return (self.job_id, job)
+        return (job_id, job)
 
-    def _insert_remaining(self, job_id: str, op_id: str) -> Tuple[str, FaasJob]:
+    def _insert_remaining(self, job_id: str, op_id: str) -> Tuple[str, dict]:
         """Insert faas job remaining execution metadata into FireStore `job.remaining` collection
 
         Args:
@@ -350,6 +293,12 @@ class FaasJobManager(metaclass=Singleton):
         self.firestore.upsert(self.remaining_job_collection_name, op_id, {}, doc_ref)
         return (self.op_id, {})
 
+    def _insert_job_done(self, job_done_collection: str, job_args: dict, status: FaasOpState) -> Tuple[str, dict]:
+        data = {"args": job_args, "status": status.value}
+        self.firestore.upsert(job_done_collection, self.job_id, data, None)
+        return (self.op_id, data)
+
+
     def _delete_remaining(self, job_id: str, op_id: Optional[str]) -> bool:
         """Deletes metadata of remaining job to be executed
 
@@ -366,7 +315,7 @@ class FaasJobManager(metaclass=Singleton):
             remaining_ref: DocumentReference = doc_ref.collection(
                 self.remaining_job_collection_name
             ).document(op_id)
-            return self.firestore.delete_document(remaining_ref)
+            self.firestore.delete_document(remaining_ref)
         return True
 
     def _insert_failed(
@@ -416,25 +365,37 @@ class FaasJobManager(metaclass=Singleton):
         """
 
         _job_state = FaasOpState.CRTD
-        err_callback = None
         _err_exception = None
 
         try:
             if req.job_id is None:
                 self.job_id = (str)(uuid.uuid4())
                 self.op_id = None
-                self._upsert_job(name=job_name, args=req)
+                self.user_id = req.user_id
+                self.username = req.username
+                self._upsert_job(name=job_name, job_id=self.job_id, args=req)
                 self.job_parent_idx_list = []
+                self.job_done_collection = req.job_done_collection
             else:
                 self.job_id = req.job_id
                 self.op_id = req.op_id
+                self.user_id = req.user_id
+                self.username = req.username
+                self.job_parent_idx_list = req.job_child_idx_list
+                self.job_done_collection = req.job_done_collection
             yield self
         except Exception as ex:
             _err_exception = ex
             _job_state = FaasOpState.ERR
-            err_callback = get_err_callback(req)
         finally:
-            self._job_close(job_name, req, _job_state, _err_exception, err_callback)
+            self._job_close(job_name, req, _job_state, _err_exception)
+
+    def finish_job(self, req: Request, root_job: FaasJob, status: FaasOpState):
+        """Save a job done information in a firestore collection
+        """
+
+        if req.job_done_collection is not None:
+            self._insert_job_done(req.job_done_collection, root_job.args, status)
 
     def add_job(self, trigger: FaasJobTrigger):
         """Increments total jobs counter on FaasJob metadata, add
@@ -447,8 +408,13 @@ class FaasJobManager(metaclass=Singleton):
 
         new_idx_list = self.job_parent_idx_list.copy()
         new_idx_list.append(self.new_jobs_cnt)
+        trigger.message["job_id"] = self.job_id
         trigger.message["op_id"] = (str)(uuid.uuid4())
         trigger.message["job_child_idx_list"] = new_idx_list
+        trigger.message["user_id"] = self.user_id
+        trigger.message["username"] = self.username
+        if self.job_done_collection is not None:
+            trigger.message["job_done_collection"] = self.job_done_collection
         self.new_jobs_cnt += 1
         self._insert_remaining(self.job_id, trigger.message["op_id"])
         job_ref = self.firestore.get_document_ref(
@@ -457,7 +423,7 @@ class FaasJobManager(metaclass=Singleton):
             shard_collection_name=self.job_collection_name,
             shard_idx_list=self.job_parent_idx_list,
         )
-        trigger._collection: CollectionReference = job_ref.collection(
+        trigger._collection = job_ref.collection(
             self.job_collection_name
         )
         trigger._job = self._get_or_default_job(
